/*
 * Copyright (C) 2016 Sebastian Geiger
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef GF_CANDIDATE_BOX_H
#define GF_CANDIDATE_BOX_H

#include <gtk/gtk.h>

G_BEGIN_DECLS

#define GF_TYPE_CANDIDATE_BOX gf_candidate_box_get_type ()
G_DECLARE_FINAL_TYPE (GfCandidateBox, gf_candidate_box,
                      GF, CANDIDATE_BOX, GtkEventBox)

GtkWidget *gf_candidate_box_new          (guint           index);

void       gf_candidate_box_set_labels   (GfCandidateBox *box,
                                          const gchar    *index_label,
                                          const gchar    *candidate_label);

guint      gf_candidate_box_get_index    (GfCandidateBox *box);

void       gf_candidate_box_set_selected (GfCandidateBox *box,
                                          gboolean        selected);

G_END_DECLS

#endif
